// Grab and Drag 0.7.3 extension by Ian Weiner: ian.weiner@gmail.com
//
// See license.txt for more information

const gadVERSION = "0.7.3";
const gadFF = 0; 
const gadTB = 1;
const gadSTDCURSOR = "move";
const gadDRAGCURSOR = "move";

const gadPrefListener = {
  domain: "grabAndDrag.",
  observe: function(subject, topic, prefName) {
    gadSync();
  }
}

var gadContentArea, gadRenderingArea, gadAppContent, gadMainWin;
var gadScrollObj;
var gadStartY, gadStartX;
var gadOn, gadVer, gadFirstTime=false;
var gadInF2=false, gadMoved=false, gadAppPlatform;

var gadPrefService = Components.classes["@mozilla.org/preferences-service;1"]
                     .getService(Components.interfaces.nsIPrefService);
var gadPref = gadPrefService.getBranch("grabAndDrag.");
var gadPrefRoot = gadPrefService.getBranch(null);
var gadPbi = gadPrefRoot.QueryInterface(Components.interfaces.nsIPrefBranchInternal);
gadPbi.addObserver(gadPrefListener.domain, gadPrefListener, false);

function gadInit() {  
  
  try {
    gadOn = gadPref.getBoolPref("on");
  }
  catch(err) { gadPref.setBoolPref("on", "false"); gadOn=false; } 
  try {
    gadVer = gadPref.getCharPref("version");
  }
  catch(err) { gadPref.setCharPref("version", gadVERSION); gadVer = gadVERSION; gadFirstTime=true; } 
  if (gadVer != gadVERSION) {
    gadPref.setCharPref("version", gadVERSION); 
    gadFirstTime=true; 
  }

  // determine which app we're running under, firefox or thunderbird
  if (document.getElementById("main-window")) {
    gadAppPlatform=gadFF; 
    gadMainWin = document.getElementById("main-window");
    gadContentArea = document.getElementById("content");
    gadRenderingArea = gadContentArea.mPanelContainer;
    gadAppContent = document.getElementById("appcontent");
  } else { 
    gadAppPlatform=gadTB; 
    gadMainWin = document.getElementById("messengerWindow");
    gadContentArea = document.getElementById("messagepane");
    gadRenderingArea = document.getElementById("messagepane");
    gadAppContent = document.getElementById("messengerWindow");
  }

  if (gadOn) {
  	gadRenderingArea.addEventListener("mousedown", gadDownHandler, true);
	gadRenderingArea.addEventListener("click",gadClickHandler, true);
    gadAppContent.addEventListener("DOMContentLoaded", gadLoad, true);
	gadAppContent.addEventListener("click",gadClickPreempt, true);
	window.addEventListener("keydown", gadKeyDownHandler, false);
	window.addEventListener("keyup", gadKeyUpHandler, false);
  }  

  try { updateButtonImage(); } catch(err) { }
}

//preempt the normal image size toggling to allow dragging of images
function gadClickPreempt(e) { 
	if (e.originalTarget.ownerDocument.toggleImageSize && e.button==0) { 
		e.preventDefault(); e.stopPropagation(); return false;
	}
}

function gadSync() {
	if (gadOn != gadPref.getBoolPref("on"))  gadToggle();
}

function gadKeyDownHandler(e) { 	
	if (e.keyCode==e.DOM_VK_F2 && !gadInF2) { 
		gadSetCursor(e.target.ownerDocument, "auto"); 
		gadInF2=true; 
		window.setCursor("auto");
		e.preventDefault(); e.stopPropagation(); return false;
	}
}

function gadKeyUpHandler(e) { 
	if (e.keyCode==e.DOM_VK_F2) { 
		gadInF2=false; 
		gadSetCursor(e.target.ownerDocument, gadSTDCURSOR); 
		e.preventDefault(); e.stopPropagation(); return false;
	}
}


function gadSetCursor(doc, cur) {
	if (!doc) return;
    var i;		
    if (doc.documentElement.nodeName.toLowerCase() == "html") {
	  if (doc.evaluate) {
	  	var x = doc.evaluate("//body", doc, null, XPathResult.UNORDERED_NODE_SNAPSHOT_TYPE,null);
	  	if (x) {
	     	for (i=0; i<x.snapshotLength; i++) {
    	     	try { x.snapshotItem(i).style.removeProperty("cursor"); } catch(err) { }
    	     	x.snapshotItem(i).style.setProperty("cursor", cur, "important");     	 
         	}
      	}
      	x = doc.evaluate("//frame | //iframe", doc, null, XPathResult.UNORDERED_NODE_SNAPSHOT_TYPE,null);
	  	if (x) {
        	for (i=0; i<x.snapshotLength; i++) {
		    	gadSetCursor(x.snapshotItem(i).contentDocument,cur);
         	}
      	}
	  } else {
    	try { doc.body.style.removeProperty("cursor"); } catch(err) { }
    	doc.body.style.setProperty("cursor", cur, "important");     	 
	  }
    } else {
		// to be implemented?
	}
}

function gadLoad(e) {
	if (gadFirstTime) {
		gadFirstTime=false;
		alert("Thank you for trying Grab and Drag 0.7.3. This message will only appear once. Please note the following: \n" +
			  "\n" +
              "* You can install a toggle icon via the menus View > Toolbars > Customize...\n" +
			  "* There is a hotkey toggle as well: ALT-SHIFT-D\n" +
			  "* There is a 'quick toggle': Holding down F2 disables Grab and Drag momentarily, allowing you to select text quickly.");
	}
	gadSetCursor(e.target, gadSTDCURSOR);
}

function updateButtonImage() {
	var b = document.getElementById("grabanddrag-button");
	var bigsmall = "";
	try { bigsmall = ((b.parentNode.getAttribute("iconsize")=="small")?"small":""); } catch(err) { }
	try { b.style.removeProperty("list-style-image"); } catch(err) { }
	b.style.setProperty("list-style-image", "url(chrome://grabanddrag/skin/grabanddrag" + (gadOn?"":"off") + bigsmall + ".png)", "important");	
}

function gadToggle() {
   var b;
	
   gadOn=!gadOn; 
   try { gadPref.setBoolPref("on", gadOn); } catch(err) { }
   try { updateButtonImage(); } catch(err) { }
   if (gadAppPlatform==gadFF) {
     for (var i=0; b = gadContentArea.browsers[i]; i++) {
       try { gadSetCursor(b.contentDocument,(gadOn?gadSTDCURSOR:"auto")); } catch(err) { }
     }
   } else { 
     try { gadSetCursor(gadRenderingArea.contentDocument,(gadOn?gadSTDCURSOR:"auto")); } catch(err) { }
   }
   if (gadOn) { 
	   	window.goDoCommand('cmd_selectNone'); 
  	   	gadRenderingArea.addEventListener("mousedown", gadDownHandler, true);
		gadRenderingArea.addEventListener("click",gadClickHandler, true);
       	gadAppContent.addEventListener("DOMContentLoaded", gadLoad, true);
		gadAppContent.addEventListener("click",gadClickPreempt, true);
	   	window.addEventListener("keydown", gadKeyDownHandler, false);
       	window.addEventListener("keyup", gadKeyUpHandler, false);
   } else {	   
  	   	try { gadRenderingArea.removeEventListener("mousedown", gadDownHandler, true); } catch(err) { }
  	   	try { gadRenderingArea.removeEventListener("click",gadClickHandler, true); } catch(err) { }
       	try { gadAppContent.removeEventListener("DOMContentLoaded", gadLoad, true); } catch(err) { }
		try { gadContentArea.removeEventListener("dragover", gadUpHandler, false); } catch(err) { }
		try { gadAppContent.removeEventListener("click",gadClickPreempt, true); } catch(err) { }
	   	try { window.removeEventListener("mouseup", gadUpHandler, true); } catch(err) { }
   	   	try { window.removeEventListener("mousedown", gadUpHandler, true); } catch(err) { }
       	try { window.removeEventListener("mousemove", gadMoveScrollBar, true); } catch(err) { }
	   	try { window.removeEventListener("keydown", gadKeyDownHandler, false); } catch(err) { }
	   	try { window.removeEventListener("keyup", gadKeyUpHandler, false); } catch(err) { }
   }
}


function gadOnItem(initialNode) {
  var nextNode, currNode;
  var doc = initialNode.ownerDocument;
  var st = doc.defaultView.getComputedStyle(initialNode, "");
  var cur = st.getPropertyValue("cursor");

  //if (!gadContentArea.contentDocument.isSameNode(doc)) return true;
  if (doc.imageResizingEnabled) return false;
  if (initialNode.nodeName.toLowerCase() == "textarea") 
    return !(initialNode.getAttribute("disabled") || initialNode.getAttribute("readonly"));
  if (st.getPropertyValue("-moz-user-select")=="none") return true;
  if ((cur != gadSTDCURSOR) && (cur != "auto")) return true;
  
  nextNode = initialNode; 
  do {
        currNode = nextNode;
        try { 
	       if ((currNode.onclick) || (currNode.ondrag) || (currNode.onmousedown) || 
               (currNode.onmouseup) || (currNode.ondblclick)) 
              return true; 
           if (currNode.hasAttribute("href")) return true; 
	       if ((currNode.nodeName.toLowerCase() == "scrollbar") ||
      		   (currNode.nodeName.toLowerCase() == "input") ||
               (currNode.nodeName.toLowerCase() == "select") ||
      		   (currNode.nodeName.toLowerCase() == "option") ||
      		   (currNode.nodeName.toLowerCase() == "embed") ||
      		   (currNode.nodeName.toLowerCase() == "tree") ||
      		   (currNode.nodeName.toLowerCase() == "applet") ||
      		   (currNode.nodeName.toLowerCase() == "statusbar")) return true; 
        } catch(err) { }
        nextNode = currNode.parentNode;
  } while (nextNode && currNode != doc.documentElement);
  
  return false;
}

function gadDownHandler(e) {  
  	var t = e.originalTarget;
  	if (e.button!=0 || gadOnItem(t) || gadInF2) return true; 
  	e.preventDefault();    
  	gadHasMoved = false;
  	gadContentArea.addEventListener("dragover", gadUpHandler, false);
  	gadScrollObj = gadFindNodeToScroll(t);
  	window.addEventListener("mouseup", gadUpHandler, true);
  	window.addEventListener("mousemove", gadMoveScrollBar, true);
  	if (gadScrollObj.scrollType == 3) return false;
  	gadRenderingArea.removeEventListener("mousedown", gadDownHandler, true);
  	window.addEventListener("mousedown", gadUpHandler, true);
  	gadStartY = e.screenY; gadStartX = e.screenX;
  	return false;
}

function gadMoveScrollBar(e) {
   e.preventDefault(); e.stopPropagation(); 
   window.setCursor(gadDRAGCURSOR);
   var dX = gadStartX - e.screenX; var dY = gadStartY - e.screenY;
   if (Math.abs(dX) > 2 || Math.abs(dY) > 2) gadHasMoved=true;
   gadStartY = e.screenY; gadStartX = e.screenX;
   if (gadScrollObj.isXML)
    switch (gadScrollObj.scrollType) {
     case 1: gadScrollObj.clientFrame.scrollBy(0, Math.ceil(dY));
             break;
     case 2: gadScrollObj.clientFrame.scrollBy(Math.ceil(dX), 0);
             break;
     case 0: if (Math.abs(dX) > Math.abs(dY)) {
                  gadScrollObj.clientFrame.scrollBy(Math.ceil(dX), 0);
             } else { 
	              gadScrollObj.clientFrame.scrollBy(0, Math.ceil(dY));
             }
   }
   else
    switch (gadScrollObj.scrollType) {
     case 1: gadScrollObj.nodeToScroll.scrollTop += Math.ceil(dY);
             break;
     case 2: gadScrollObj.nodeToScroll.scrollLeft += Math.ceil(dX);
             break;
     case 0: if (Math.abs(dX) > Math.abs(dY)) {
                  gadScrollObj.nodeToScroll.scrollLeft += Math.ceil(dX);
             } else {
	              gadScrollObj.nodeToScroll.scrollTop += Math.ceil(dY);
	         }
   }
}

function gadUpHandler(e) {
   	window.setCursor("auto");	
   	window.removeEventListener("mouseup", gadUpHandler, true);
   	window.removeEventListener("mousedown", gadUpHandler, true);
   	window.removeEventListener("mousemove", gadMoveScrollBar, true);
   	gadContentArea.removeEventListener("dragover", gadUpHandler, false);
   	e.preventDefault();       
   	gadRenderingArea.addEventListener("mousedown", gadDownHandler, true);
   	if (!gadHasMoved && (e.originalTarget.ownerDocument.toggleImageSize))
		e.originalTarget.ownerDocument.toggleImageSize(); 
}

function gadFixFocus(n) {
	var fr;
	var focused = document.commandDispatcher.focusedElement;
	if (focused) focused.blur();
	content.focus();
	try { fr = n.ownerDocument.defaultView.frameElement.contentWindow; } catch(err) { }
	if (fr) fr.focus();
}

//handle some useful default browser click actions that we prevented with preventDefault 
//above.
function gadClickHandler(e) {
	var t = e.originalTarget;
	if (gadOnItem(t) || e.button != 0) return; 
 	if (gadInF2) { e.preventDefault(); return; }
 	gadFixFocus(e.originalTarget);
	window.goDoCommand('cmd_selectNone'); 
}

// The following is only slightly modified from Marc Boullet's All-in-one Gestures extension
function gadFindNodeToScroll(initialNode) {

  function getStyle(elem, aProp) {
    var p = elem.ownerDocument.defaultView.getComputedStyle(elem, "").getPropertyValue(aProp);
    var val = parseFloat(p);
    if (!isNaN(val)) return Math.ceil(val);
    if (p == "thin") return 1;
    if (p == "medium") return 3;
    if (p == "thick") return 5;
    return 0;
  }

  function scrollCursorType(neededW, availW, neededH, availH, scrollBarSize) {
    if (neededW <= availW && neededH <= availH) return 3;
    if (neededW > availW && neededH > availH) return 0;
    if (neededW > availW) return ((neededH <= (availH - scrollBarSize)) - 0) << 1;  // 0 or 2
    return (neededW <= (availW - scrollBarSize)) - 0;
  }
  
  const defaultScrollBarSize = 16;
  const twiceScrollBarSize = defaultScrollBarSize * 2;
  var retObj = {scrollType: 3, isXML: false, nodeToScroll: null, 
                clientFrame: null, isBody: false, isFrame: false,
                targetDoc: null, insertionNode: null};
  var nextNode, currNode;
  var targetDoc = initialNode.ownerDocument;
  var docEl = targetDoc.documentElement;
  retObj.insertionNode = (docEl) ? docEl : targetDoc;
  var docBox = targetDoc.getBoxObjectFor(retObj.insertionNode);
  var clientFrame = targetDoc.defaultView;
  retObj.targetDoc = targetDoc; retObj.clientFrame = clientFrame;
  if (docEl && docEl.nodeName.toLowerCase() == "html") { // walk the tree up looking for something to scroll
     if (clientFrame.frameElement) retObj.isFrame = true; else retObj.isFrame = false;
     var bodies = docEl.getElementsByTagName("body");
     if (!bodies || !bodies.length) return retObj;
     var bodyEl = bodies[0];
     if (initialNode == docEl) nextNode = bodyEl;
     else if (initialNode.nodeName.toLowerCase() == "select") nextNode = initialNode.parentNode;
          else nextNode = initialNode;
     do {
        try {
          currNode = nextNode;
          if (currNode.clientWidth && currNode.clientHeight &&
             currNode.ownerDocument.defaultView.getComputedStyle(currNode, "").getPropertyValue("overflow") != "hidden") {
             realWidth = currNode.clientWidth + getStyle(currNode, "border-left-width") + getStyle(currNode, "border-right-width");
             realHeight = currNode.clientHeight + getStyle(currNode, "border-top-width") + getStyle(currNode, "border-bottom-width");
             retObj.scrollType = scrollCursorType(currNode.scrollWidth, realWidth, currNode.scrollHeight, realHeight, 0);
             if (retObj.scrollType != 3) {
                retObj.nodeToScroll = currNode;
                retObj.isBody = currNode == bodyEl || currNode == docEl;
                return retObj;
             }
          }
          nextNode = currNode.parentNode;
        }
        catch(err) {return retObj;}
     } while (nextNode && currNode != docEl);
     if (retObj.isFrame) return gadFindNodeToScroll(clientFrame.frameElement.ownerDocument.documentElement);
  }
  else { // XML document; do our best
     retObj.nodeToScroll = initialNode;
     if (docBox) {
        try { retObj.scrollType = scrollCursorType(docBox.width, gadRenderingArea.boxObject.width,
                   docBox.height, gadRenderingArea.boxObject.height, defaultScrollBarSize); } catch(err) { }
     }
     retObj.isXML = true;
  }
  return retObj;
}
